"""overpass and osm download functions"""
import os
import pathlib
import tempfile
import urllib
from urllib.request import urlopen


def dl_osm_from_overpass(area, bbox):
    if not area and not bbox:
        raise Exception('At lease area or bbox must be given.')

    overpass_query = build_overpass_query(area, bbox)
    overpass_api_url = "http://overpass-api.de/api/interpreter"

    filepath = tempfile.mktemp(suffix='_overpass.osm')
    resp = urlopen(overpass_api_url, overpass_query.encode('utf-8'))

    if resp.status == 200:
        with open(filepath, 'w') as osm:
            osm.write(resp.read().decode('utf-8'))
        return os.path.split(filepath)[-1], filepath
    else:
        raise urllib.request.HTTPError(
            'Error calling Overpass API. Status: {}, reason: {}'.format(
                resp.status, resp.reason))

    raise Exception("Can't download data form overpass api.")


def build_overpass_query(area, bbox):
    template = """
    {bbox}
    {area}
    (
      rel
        [!"deleted"]
        ["type"="route"]
        ["route"~"tram|subway|bus|ex-bus|light_rail|rail|railway"]
        {area_limit};
      rel
        [!"deleted"]
        ["type"="public_transport"]
        ["public_transport"="stop_area"]
        {area_limit};
    );
    out body;

    way(r){area_limit};
    out body;

    node(w){area_limit};
    out body;
    """
    bbox_fmt = ''
    area_fmt = ''
    area_limit_fmt = ''

    if bbox:
        south, west, north, east = bbox.split(',')
        bbox_fmt = '[bbox:{},{},{},{}];'.format(south, west, north, east)

    if area:
        area_fmt = 'area["name"="{}"]->.searchArea;'.format(area)
        area_limit_fmt = '(area.searchArea)'

    return template.format(
        bbox=bbox_fmt, area=area_fmt, area_limit=area_limit_fmt)


def dl_osm_from_url(url):
    filename = tempfile.mktemp(suffix=pathlib.Path(url).name)
    filepath, headers =\
        urllib.request.urlretrieve(url, filename=filename)
    return pathlib.Path(url).name, filepath



#!env python
"""
Extracts partial GTFS data from OSM file.
"""
import os
import time
import tempfile
import logging
from logging import FileHandler
from contextlib import contextmanager
from pathlib import Path

import argparse

from o2g import __version__
from o2g.gtfs import gtfs_dummy
from o2g.gtfs.gtfs_writer import GTFSWriter
from o2g.osm.exporter import TransitDataExporter
from o2g.web import dl_osm_from_overpass


class readable_dir(argparse.Action):
    def __call__(self, parser, namespace, prospective_dir, option_string=None):
        if not os.path.isdir(prospective_dir):
            parser.print_usage()
            print('Try "{} --help" for help.\n'.format(parser.prog))
            parser.exit("Error: Invalid path: {0}".format(prospective_dir))
        if os.access(prospective_dir, os.R_OK):
            setattr(namespace, self.dest, prospective_dir)
        else:
            parser.print_usage()
            print('Try "{} --help" for help.\n'.format(parser.prog))
            parser.exit("Error: Unreadable dir: {0}".format(prospective_dir))


class readable_file(argparse.Action):
    def __call__(self, parser, namespace, prospective_file, option_string=None):
        if not os.path.isfile(prospective_file):
            parser.print_usage()
            print('Try "{} --help" for help.\n'.format(parser.prog))
            parser.exit("Error: Invalid file: {0}".format(prospective_file))
        if os.access(prospective_file, os.R_OK):
            setattr(namespace, self.dest, prospective_file)
        else:
            parser.print_usage()
            print('Try "{} --help" for help.\n'.format(parser.prog))
            parser.exit("Error: Unreadable file: {0}".format(prospective_file))


def cli():
    parser = argparse.ArgumentParser(
        prog='o2g',
        epilog='Here is a smile for you :)',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Export GTFS feed from OpenStreetMap data.')
    parser.add_argument('osmfile', metavar='OSMFILE', action=readable_file,
                        nargs='?',
                        default=argparse.SUPPRESS,
                        help='an OSM data file supported by osmium')
    parser.add_argument('--area',
                        help='an OSM area name, e.g. Freiburg')
    parser.add_argument('--bbox',
                        help='a boundary box, e.g. 47.9485,7.7066,48.1161,8.0049')
    parser.add_argument('--outdir', action=readable_dir,
                        default='.',
                        help='output directory')
    parser.add_argument('--zipfile',
                        help='output zip file name')
    parser.add_argument('--dummy', action='store_true',
                        default=False,
                        help='fill the missing parts with dummy data')
    parser.add_argument('--loglevel',
                        default='WARNING',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        help='the logging level')
    parser.add_argument('--version',
                        action='version',
                        version='%(prog)s ' + __version__,
                        help='show the version and exit')

    args = parser.parse_args()
    if not args.area and not args.bbox and not hasattr(args, 'osmfile'):
        parser.print_usage()
        parser.exit("o2g: error: one of these args are required: OSMFILE, bbox or area.")

    if args.loglevel:
        logging.basicConfig(level=args.loglevel)
    if hasattr(args, 'osmfile'):
        logging.debug('Input: %s', args.osmfile)
    logging.debug('Area: %s', args.area)
    logging.debug('Boundary box: %s', args.bbox)
    logging.debug('Output: %s', args.outdir)
    logging.debug('Zip?: %s', args.zipfile or False)
    logging.debug('Dummy?: %s', args.dummy)

    if args.area or args.bbox:
        filename, filepath = dl_osm_from_overpass(args.area, args.bbox)
        osmfile = filepath
    else:
        osmfile = args.osmfile

    main(osmfile, args.outdir, args.zipfile, args.dummy)


def main(osmfile, outdir, zipfile, dummy):
    start = time.time()

    with capture_logs() as logfile:
        tde = TransitDataExporter(osmfile)
        tde.process()
        logging.debug('Preprocessing took %d seconds.', (time.time() - start))

        writer = GTFSWriter()
        patched_agencies = None
        if dummy:
            dummy_data = gtfs_dummy.create_dummy_data(list(tde.routes),
                                                      list(tde.stops))
            writer.add_trips(dummy_data.trips)
            writer.add_stop_times(dummy_data.stop_times)
            writer.add_calendar(dummy_data.calendar)
            writer.add_frequencies(dummy_data.frequencies)
            patched_agencies = gtfs_dummy.patch_agencies(tde.agencies)

        if patched_agencies:
            writer.add_agencies(patched_agencies)
        else:
            writer.add_agencies(tde.agencies)
        writer.add_stops(tde.stops)
        writer.add_routes(tde.routes)
        writer.add_shapes(tde.shapes)
        writer.add_feedinfo({
            'feed_publisher_name': 'Generated by o2g',
            'feed_publisher_url': 'hiposfer.com',
            'feed_lang': 'en',
            'feed_version': int(time.time())
        })

        writer.add_file('LICENSE', Path(__file__).parents[0] / 'ODbL-1.0.txt')
        writer.add_file('logs.txt', logfile)

    if zipfile:
        writer.write_zipped(os.path.join(outdir, zipfile))
        logging.info('GTFS feed saved in %s' % os.path.join(outdir, zipfile))
    else:
        writer.write_unzipped(outdir)
        logging.info('GTFS feed saved in %s' % outdir)

    logging.debug('Done in %d seconds.', (time.time() - start))


@contextmanager
def capture_logs():
    logfile = tempfile.mktemp()
    handler = FileHandler(logfile, mode='w')
    logging.getLogger().addHandler(handler)
    try:
        yield logfile
    finally:
        logging.getLogger().removeHandler(handler)


if __name__ == '__main__':
    cli()



"""Functionality to build a list of stops."""
import logging

from o2g.osm.models import Stop


def build_stops(relations, nodes):
    visited_stops_ids = set()
    stop_to_station_map = {}

    # First gsocess all stop_areas
    for rel in relations.values():
        if rel.tags.get('public_transport') != 'stop_area':
            continue

        station = build_parent_stop(rel, nodes)
        if not station:
            continue

        for _, member_id, _ in rel.member_info:
            stop_to_station_map[member_id] = station.stop_id

        visited_stops_ids.add(station.stop_id)
        yield station


    for rel in relations.values():
        if rel.tags.get('public_transport') == 'stop_area':
            continue

        for stop in extract_stops(rel, nodes, visited_stops_ids, stop_to_station_map):
            if stop:
                yield stop


def build_parent_stop(relation, nodes):
    # One stop per stop_area is necessary.
    station_node = None
    some_node = None
    for member_type, member_id, member_role in relation.member_info:
        if member_id not in nodes:
            continue
        some_node = nodes[member_id]
        if not some_node:
            raise Exception()
        if some_node.tags.get('public_transport') == 'station' or \
            some_node.tags.get('amenity') == 'bus_station':
            station_node = some_node

    if not station_node and not some_node:
        logging.debug('No parent node found: https://osm.org/relation/%s' % relation.id)
        return

    if not station_node:
        logging.warning(
            'stop_area without station: https://osm.org/relation/%s.'\
            ' Using node https://osm.org/node/%s instead.' % (relation.id, some_node.id))
        station_node = some_node

    return Stop(
        station_node.id,
        station_node.tags.get('name') or "Unnamed stop_area.",
        station_node.lon,
        station_node.lat,
        None,
        _map_wheelchair(station_node.tags.get('wheelchair')),
        1,  # A station in GTFS terms
        '')  # Blank values since stations can't contain other stations.


def extract_stops(relation, nodes, visited_stop_ids, stop_to_station_map):
    """Extract stops in a relation."""
    # member_role: stop, halt, platform, terminal, etc.
    for member_type, member_id, member_role in relation.member_info:

        if member_id not in visited_stop_ids and \
            member_id in nodes and\
                member_role in ('stop', 'halt'):

            location_type = ''

            visited_stop_ids.add(member_id)
            yield Stop(
                member_id,
                nodes[member_id].tags.get('name') or
                "Unnamed {} stop.".format(relation.tags.get('route')),
                nodes[member_id].lon if member_id in nodes else '',
                nodes[member_id].lat if member_id in nodes else '',
                relation.id,
                _map_wheelchair(nodes[member_id].tags.get('wheelchair')),
                location_type,
                stop_to_station_map.get(member_id, ''))


def _map_wheelchair(osm_value):
    if not osm_value:
        return 0
    elif osm_value == 'limited':
        return 1
    elif osm_value == 'yes':
        return 1
    elif osm_value == 'no':
        return 2
    else:
        logging.warning('Unknown OSM wheelchair value %s', osm_value)
        return 0



"""o2g web interface"""
import os
import tempfile
from urllib.parse import urlparse

from bottle import run, template, request, static_file, abort, default_app

from o2g.cli import main
from o2g.web import dl_osm_from_overpass, dl_osm_from_url

app = default_app()


@app.get('/')
def index():
    return template('index.html', messages=[])


@app.post('/')
def do_index():
    file = request.files.get('file')
    url = request.forms.get('url')

    if not url and not file:
        return template('index.html', messages=['Provide a URL or upload a file.'])

    if not file and url and not is_valid_url(url):
        return template('index.html', messages=['Not a valid URL.'])

    if file:
        filename = file.filename
        filepath = save_file(file)
    elif url:
        filename, filepath = dl_osm_from_url(url)

    return send_feed(
        filepath,
        filename + '.gtfs.zip',
        dummy=bool(request.forms.get('dummy')))


def is_valid_url(url):
    return urlparse(url).scheme in ('http', 'https')


def save_file(file):
    fd, filepath = tempfile.mkstemp(suffix=file.filename)
    file.save(filepath, overwrite=True)
    return filepath


def send_feed(filepath, filename=None, dummy=False):
    zipfile = create_zipfeed(filepath, dummy)
    if not filename:
        filename = os.path.split(zipfile)[-1]

    return static_file(
        os.path.split(zipfile)[-1],
        root=os.path.split(zipfile)[0],
        download=filename,
        mimetype='application/zip')


def create_zipfeed(filename, dummy=False):
    zipfile = '{}.zip'.format(filename)
    print('Writing GTFS feed to %s' % zipfile)
    main(filename, '.', zipfile, dummy)
    return zipfile


@app.get('/o2g', methods=['GET'])
def o2g():
    area = request.params.get('area')
    bbox = request.params.get('bbox')
    url = request.params.get('url')

    if area or bbox:
        filename, filepath = dl_osm_from_overpass(area, bbox)
        if area:
            filename = area + '_overpass'
    elif url and is_valid_url(url):
        filename, filepath = dl_osm_from_url(url)
    else:
        abort(400)

    return send_feed(
        filepath,
        filename + '.gtfs.zip',
        dummy=bool(request.params.get('dummy')))


if __name__ == '__main__':
    run(host='0.0.0.0', port=int(os.getenv('PORT', 3000)), debug=True)



import logging

import osmium as o

from o2g.osm.models import Node


class NodeHandler(o.SimpleHandler):
    def __init__(self, node_ids):
        super(NodeHandler, self).__init__()
        self.node_ids = node_ids
        self.nodes = {}

    @property
    def missing_node_ids(self):
        """Get a list of nodes not found in OSM data."""
        present_node_ids = self.nodes.keys()
        for nid in self.node_ids:
            if nid not in present_node_ids:
                yield nid

    def node(self, n):
        """Process each node."""
        if n.id not in self.node_ids:
            return

        try:
            self.nodes[n.id] =\
              Node(n.id,
                   n.location.lon,
                   n.location.lat,
                   {t.k: t.v for t in n.tags})
        except o.InvalidLocationError:
            logging.debug('InvalidLocationError at node %s', n.id)
